import os

from typing import List, Dict

from PySide6.QtCore import QAbstractListModel, Qt, QModelIndex
from PySide6.QtGui import QIcon


class bookmarks_model(QAbstractListModel):
    """
    Handles the bookmarks as file and as list of dicts.
    """

    def __init__(self, *args, path_to_bookmark_file: str, **kwargs):
        """
        Reads bookmarks from file to internal list of dicts or creates the
        file if it doesn't exist yet. The file should have the format of one
        bookmark per line and the individual lines should have the format:

        bookmark name|/path/to/bookmark

        :param path_to_bookmark_file: Path to the file, in which the bookmarks
                                      are saved
        :type path_to_bookmark_file: str
        """
        super(bookmarks_model, self).__init__(*args, **kwargs)
        self.list = []
        if os.path.exists(path_to_bookmark_file):
            self.list = self.get_bookmarks_from_file(path_to_bookmark_file)
        else:
            open(path_to_bookmark_file, 'a').close()
        self.path_to_bookmark_file = path_to_bookmark_file

    def data(self, index: QModelIndex, role: int):
        """
        Returns the data of the object in a Qt-conforming way. If the given
        role is DisplayRole it returns the name of the device as a string, if
        it's DecorationRole it returns an icon according to it's mount state.

        :param index: Index of the item in the data structure.
        :type index: QModelIndex
        :param role: Item role according to Qt.
        :type role: int
        :return: Device name or icon according to mount state.
        """
        if role == Qt.DisplayRole:
            return (self.list[index.row()])['name']
        if role == Qt.DecorationRole:
            return QIcon.fromTheme('user-bookmarks')

    def rowCount(self, index: int) -> int:
        """
        Returns the number of mountable devices in a Qt-conforming way.

        :param index: unused, but needed by Qt.
        :type index: int
        :return: Number of mountable devices.
        :rtype: int
        """
        return len(self.list)

    def get_bookmarks_from_file(self, path: str) -> List[Dict[str, str]]:
        """
        Gets bookmarks from a file.

        :raises OSError: If the file, the bookmarks are saved in, can't be
                         read.
        :param path: Path to the file, in which the bookmarks are saved.
        :type path: str
        :return: List of dicts containing name and path of the bookmarks.
                 The dicts follow this structure: {'name': '', 'path': ''}
        :rtype: List[Dict[str, str]]
        """
        list = []
        with open(path) as bookmarks:
            for bookmark in bookmarks:
                list.append({'name': bookmark.split('|')[0],
                             'path': bookmark.split('|')[1].rstrip()})
        return list

    def add_bookmark(self, name: str, path: str):
        """
        Adds a new bookmark. Does nothing, if another bookmark with the same
        name already exists or the | character is in the chosen name. The
        bookmark will be saved to the bookmark file that was chosen on
        instantiation of this class.

        :raises NameError: If the given name already exists or contains the
                           '|' special character.
        :param name: The name of the new bookmark. This will be shown in the
                     UI.
        :type name: str
        :param path: Path which the bookmark should lead to.
        :type path: str
        """
        if not self.bookmark_exists(name) and '|' not in name:
            self.list.append({'name': name, 'path': path})
            with open(self.path_to_bookmark_file, 'a') as bookmarks:
                bookmarks.write(name + '|' + path + '\n')
        else:
            raise NameError

    def remove_bookmark(self, name: str):
        """
        Removes a bookmark from the internal bookmark list. It also rewrites
        the updated list to the bookmark file chosen on class instatiation.

        :param name: The name of the bookmark that should be deleted.
        :type name: str
        """
        new_list = []
        for bookmark in self.list:
            if bookmark['name'] != name:
                new_list.append(bookmark)
        self.list = new_list
        with open(self.path_to_bookmark_file, 'w') as bookmarks:
            for bookmark in self.list:
                bookmarks.write(bookmark['name']
                                + '|'
                                + bookmark['path']
                                + '\n')

    def get_path_from_name(self, name: str):
        """
        Returns the path that is registered with the given bookmark name. If
        the name doesn't exist, it returns False.

        :param name: The of the bookmark to which the path is requested.
        :type name: str
        :return: The path registered with the given bookmark name, or False,
                 if the name doesn't exist
        :rtype: str | bool
        """
        for bookmark in self.list:
            if name == bookmark['name']:
                return bookmark['path']
        return False

    def bookmark_exists(self, name: str) -> bool:
        """
        Returns True if the bookmark with the given name exists and False, if
        it doesn't.

        :param name: Name of the bookmark to check for.
        :type name: str
        :return: True if bookmark exists, False if it doesn't.
        :rtype: bool
        """
        for bookmark in self.list:
            if name == bookmark['name']:
                return True
        return False

    def get_all(self) -> List[Dict[str, str]]:
        """
        Returns all bookmarks as a list of dicts.

        :return: List of dicts of all bookmarks. The dicts follows the format
                 {'name': '', 'path': ''}
        :rtype: List[Dict[str, str]]
        """
        return self.list



#! /usr/bin/env python3

import sys

from PySide6.QtWidgets import QApplication

from tfm import tfm

if __name__ == "__main__":
    app = QApplication()
    # TODO: create clean CLI
    # TODO: document CLI
    window = tfm.tfm(sys.argv)
    window.show()
    sys.exit(app.exec())



from collections import deque


class stack:
    """
    Provides a classic stack using a deque() data structure.
    """

    def __init__(self):
        """
        Initializes the stack.
        """
        self.stack = deque()

    def empty(self) -> bool:
        """
        Returns True if stack is emtpy and False if it isn't.

        :return: Emptyness of the stack.
        :rtype: bool
        """
        return len(self.stack) == 0

    def size(self) -> int:
        """
        Returns the size of the stack.

        :return: Number of elements in the stack.
        :rtype: int
        """
        return len(self.stack)

    def top(self):
        """
        Returns the top element of the stack or None, if stack is empty.

        :return: Top element of the stack or None, if stack is empty.
        """
        if len(self.stack) > 0:
            return self.stack[-1]
        return None

    def push(self, stack_element):
        """
        Adds an element to the top of the stack.

        :param stack_element: The element which will be added to the stack.
        """
        self.stack.append(stack_element)

    def pop(self):
        """
        Removes and returns the top element of the stack. If stack is empty,
        return None.

        :return: Top element of the stack or None, if stack is empty.
        """
        if len(self.stack) > 0:
            return self.stack.pop()
        return None

    def drop(self):
        """
        Resets the entire stack by removing all of its elements.
        """
        while (self.size() != 0):
            self.stack.pop()



import os
import collections
from typing import List

from PySide6.QtCore import QObject, QRunnable, QDir, QFile, Signal, Slot
from PySide6.QtGui import QClipboard

import tfm.utility as utility


class paste_signals(QObject):
    """
    QObject class to hold the signals for the paste worker, which can't do
    that, because it a subclass of QRunnable.

    Signals:
    **finished**: is emitted, when the worker is done.
    **started**: is emitted, when the worker starts and the clipboard is not
                 empty.
    **ready**: is emitted, when all file paths to paste are collected,
               contains the number of files to paste as int
    **progress**: is emitted, whenever a file is done being pasted,
                  contains the number of files pasted so far.
    **progress_message**: is emitted, whenever **progress** is emitted, but is
                          a formatted string containing the information about
                          what is happening.
    """

    finished = Signal()
    started = Signal()
    ready = Signal(int)
    progress_message = Signal(str)
    progress = Signal(int)


class paste_worker(QRunnable):
    """
    Worker class, that is used to paste files in a different thread.
    """

    def __init__(self,
                 *args,
                 clipboard: QClipboard,
                 target_path: str,
                 marked_to_cut: List[str],
                 **kwargs):
        """
        Calls QObject constructor and initializes the clipboard, the current
        path and the list of files, that are marked to cut. It also initializes
        the is_canceld property to False.

        :param clipboard: The clipboard containing the file URLs that will be
                          pasted.
        :type clipboard: QClipboard
        :param target_path: Path, where the files should be pasted to.
        :type target_path: str
        :param marked_to_cut: A list of file paths, which shold be deleted
                              after pasting.
        :type marked_to_cut: List[str]
        """
        super(paste_worker, self).__init__(*args, **kwargs)
        self.clipboard = clipboard
        self.target_path = target_path
        self.marked_to_cut = marked_to_cut
        self.signals = paste_signals()

        self.is_canceled = False

    @Slot()
    def cancel(self):
        self.is_canceled = True

    def get_paths_from_clipboard(self) -> List[str]:
        """
        Extracts file paths from file URLs in the clipboard.

        :return: List of file paths
        :rtype: List[str]
        """
        path_list = []
        for url in self.clipboard.mimeData().urls():
            if (url.isLocalFile()):
                path_list.append(url.toLocalFile())
        return path_list

    def recurse_dirs(self, path_list: List[str]) -> List[str]:
        """
        Recurses all directories in the given list of paths and adds the
        files and directories found there to the given list.

        :param path_list: List of file paths
        :type path_list: List[str]
        :return: The original list of file paths extended by the contents
                 of any directory in the original list
        :rtype: List[str]
        """
        paths_to_add = []
        for path in path_list:
            if (os.path.isdir(path)):
                paths_to_add.extend(utility.traverse_dir(path))
        path_list.extend(paths_to_add)
        return path_list

    def count_files(self, path_list: List[str]):
        """
        Counts all files (and not directories) in the give list of paths.

        :param path_list: List of file paths
        :type path_list: List[str]
        """
        file_count = 0
        for path in path_list:
            if os.path.isfile(path):
                file_count += 1
        return file_count

    def get_base_path(self, path_list: List[str]):
        """
        Determines the largest common path of the files in the given list of
        paths.

        :param path_list: List of file paths
        :type path_list: List[str]
        """
        if (len(path_list) > 1):
            base_path = os.path.commonpath(path_list)
            if base_path in path_list and base_path != '/':
                return os.path.abspath(os.path.join(base_path, '..')) + '/'
            return base_path + '/'
        return (os.path.dirname(os.path.commonpath(path_list)) + '/')

    # Handle existing files better -> skip, overwrite, rename
    @Slot()
    def run(self):
        """
        Main logic. Pastes the files from the clipboard to the target path.
        """
        if self.clipboard.mimeData().hasUrls():
            self.signals.started.emit()

            path_list = self.get_paths_from_clipboard()

            cut = (collections.Counter(path_list)
                   == collections.Counter(self.marked_to_cut))

            path_list = self.recurse_dirs(path_list)

            files_copied = 0
            maximum = self.count_files(path_list)
            self.signals.ready.emit(maximum)

            base_path = self.get_base_path(path_list)

            # copy files to new location
            for path in path_list:
                if self.is_canceled:
                    break

                new_path = os.path.join(self.target_path,
                                        path.replace(base_path, ''))
                if (os.path.isdir(path)
                        and not QDir().exists(new_path)):
                    QDir().mkpath(new_path)
                elif (QFile().exists(path)
                        and not QFile().exists(new_path)):
                    if (QFile().copy(path, new_path)):
                        files_copied += 1
                        self.signals.progress.emit(files_copied)
                        self.signals.progress_message.emit(
                            'Pasting file ' + str(files_copied) + ' of '
                            + str(maximum) + '...')
                    else:
                        raise OSError
            # removed cut files
            if cut:
                for file_path in path_list:
                    if (not QFile().remove(file_path)):
                        raise OSError
        self.signals.finished.emit()



# This file contains misc utility functions. Better structuring is planned in
# the future.
import os
from typing import List, Tuple

from PySide6.QtWidgets import QFileSystemModel, QMessageBox
from PySide6.QtCore import QDir, QFileInfo, QMimeData, QUrl, QModelIndex

from prefixed import Float


def indexes_to_paths(files_as_indexes: List[QModelIndex]) -> List[str]:
    """
    Converts the given indexes to a list of paths.

    :param files_as_indexes: List of indexes of files.
    :type files_as_indexes: List[QModelIndex]
    :return: List of paths to the given files.
    :rtype: List[str]
    """
    files_as_path = []
    for index in files_as_indexes:
        if (index.column() == 0):
            files_as_path.append(QFileSystemModel().filePath(index))
    return files_as_path


def traverse_dir(path: str) -> List[str]:
    """
    Traverses the given directory and returns all files and dirs inside as
    paths.

    :param path: Path to traverse.
    :type path: str
    :return: Paths of files and dirs under the given path.
    :rtype: List[str]
    """
    path_list = []
    for root, dirs, files in os.walk(path):
        for item in files:
            path_list.append(os.path.join(root, item))
        for item in dirs:
            path_list.append(os.path.join(root, item))
    return path_list


# TODO: retrieve partition name (/dev/xxxxxxmpn)
def part_info(path: str) -> str:
    """
    Retrieves information about the partition which the path is on.
    :param path: Path on the partition.
    :type path: str
    :return: Decription of the partition.
    :rtype: str
    """
    # get fs statistics using statvfs system call
    part_stats = os.statvfs(path)
    fs_size = '{:!.1j}B'.format(Float(part_stats.f_frsize
                                      * part_stats.f_blocks))
    fs_free = '{:!.1j}B'.format(Float(part_stats.f_frsize
                                      * part_stats.f_bfree))
    return (fs_free + ' of ' + fs_size + ' free')


# TODO: calculate dir sizes
def file_info(paths: List[str]) -> str:
    """
    Retrieves information about the given files.

    :param path: Path to the files.
    :type path: List[str]
    :return: Information of the file.
    :rtype: str
    """
    size = 0
    for path in paths:
        file = QFileInfo(path)
        if (file.isFile()):
            size += file.size()
        elif (len(paths) > 1):
            return str(len(paths)) + ' items selected'
    size = '{:!.2j}B'.format(Float(size))
    if (len(paths) > 1):
        return str(len(paths)) + ' files selected, using ' + size
    file = QFileInfo(paths[0])
    if (file.isFile()):
        return (os.path.basename(path) + ': ' + size)
    return (os.path.basename(path))


def question_dialog(msg: str) -> QMessageBox:
    """
    Creates a default question dialog box.

    :param msg: The question which the user is asked.
    :type msg: str
    :return: The created dialog.
    :rtype: QMessageBox
    """
    msg_box = QMessageBox()
    msg_box.setText(msg)
    msg_box.setStandardButtons(QMessageBox.Yes | QMessageBox.Cancel)
    msg_box.setDefaultButton(QMessageBox.Yes)
    msg_box.setIcon(QMessageBox.Question)
    return msg_box


def message_dialog(msg: str, type: QMessageBox.Icon) -> QMessageBox:
    """
    Creates a default message dialog box.

    :param msg: The message which should be communicated to the user.
    :type msg: str
    :param type: The type of the message.
    :type type: QMessageBox.Icon
    :return: The created dialog.
    :rtype: QMessageBox
    """
    msg_box = QMessageBox()
    msg_box.setText(msg)
    msg_box.setStandardButtons(QMessageBox.Ok)
    msg_box.setDefaultButton(QMessageBox.Ok)
    msg_box.setIcon(type)
    return msg_box


def get_MIME(files_as_indexes: List[QModelIndex]) -> Tuple[List[str],
                                                           List[QUrl]]:
    """
    Converts the given files to their MIME data.

    :param files_as_indexes: List of indexes of files.
    :type files_as_indexes: List[QModelIndex]
    :return: files as str list of path and the MIME data of the given files.
    :rtype: Tuple[List[str], List[QUrl]]
    """
    files_as_path = indexes_to_paths(files_as_indexes)
    file_urls = []

    for file in files_as_path:
        file_urls.append(QUrl.fromLocalFile(file))

    mime_data = QMimeData()
    mime_data.setUrls(file_urls)

    return files_as_path, mime_data


def handle_args(args: List[str]) -> str:
    """
    Handles the given arguments, determining the starting dir.

    :param args: List of arguments given to the application.
    :type args: List[str]
    :return: The directory to start the application with.
    :rtype: str
    """
    if len(args) > 1:
        if os.path.isdir(args[1]):
            return os.path.abspath(args[1])
        if QUrl(args[1]).isValid() and os.path.isdir(QUrl(args[1]).path()):
            return os.path.abspath(QUrl(args[1]).path())
    return QDir.homePath()



import os
from typing import List
from subprocess import run, PIPE

from pyudev import Device, Devices, Context

from PySide6.QtCore import QAbstractListModel, Qt, QModelIndex
from PySide6.QtGui import QIcon


class mounts_model(QAbstractListModel):
    """
    Provides a Qt-compatible data model, handling the listing and mounting of
    devices.
    """

    def __init__(self, *args, context: Context, **kwargs):
        """
        Finds available mounts using udev and the given context.

        :param context: The udev context this class is used in.
        :type context: Context
        """
        super(mounts_model, self).__init__(*args, **kwargs)
        self.context = context
        self.devices = self.get_available_mounts()

    def data(self, index: QModelIndex, role: int):
        """
        Returns the data of the object in a Qt-conforming way. If the given
        role is DisplayRole it returns the name of the device as a string, if
        it's DecorationRole it returns an icon according to it's mount state.

        :param index: Index of the item in the data structure.
        :type index: QModelIndex
        :param role: Item role according to Qt.
        :type role: int
        :return: Device name or icon according to mount state.
        """
        if role == Qt.DisplayRole:
            return self.devices[index.row()].sys_name
        if role == Qt.DecorationRole:
            if (self.get_mount_point(self.devices[index.row()]) != ''):
                return QIcon.fromTheme('media-eject')
            else:
                return QIcon.fromTheme('drive-harddisk')

    def rowCount(self, index: int) -> int:
        """
        Returns the number of mountable devices in a Qt-conforming way.

        :param index: unused, but needed by Qt.
        :type index: int
        :return: Number of mountable devices.
        :rtype: int
        """
        return len(self.devices)

    def get_device_at(self, index: int) -> Device:
        """
        Returns the device at the given index.

        :param index: Index of the wanted device.
        :type index: int
        :return: The device at the index in the model.
        :rtype: Device
        """
        if (index >= 0 and index < len(self.devices)):
            return self.devices[index]
        else:
            raise IndexError

    def add(self, device: Device):
        """
        Adds a device to the model, if it's not already a part of it.

        :param device: The device to add.
        :type device: Device
        """
        if device not in self.devices:
            self.devices.append(device)

    def remove(self, device: Device):
        """
        Removes a device out of the model, if it exists.

        :param device: The device to remove.
        :type device: Device
        """
        if device in self.devices:
            self.devices.remove(device)

    def get_mount_point(self, device: Device) -> str:
        """
        Looks for a mount point of the given device, using the findmnt
        command from util-linux. Returns an empty string, if there is none.

        :param device: The device, of which the mount point is needed.
        :type device: Device
        :return: The mount point or an empty string, if it doesn't exists
        :rtype: str
        """
        return run(['findmnt',
                    '-n',
                    '-o',
                    'target',
                    '/dev/' + device.sys_name],
                   text=True,
                   stdout=PIPE,
                   stderr=PIPE,
                   universal_newlines=True).stdout.strip()

    def toggle_mount(self, device: Device):
        """
        Mounts an unmounted device or unmounts a mounted device using the
        udisksctl command of udisks2.

        :param device: The device to mount or unmount.
        :type device: Device
        """
        mount_point = self.get_mount_point(device)
        if (mount_point != ''):

            run(['udisksctl', 'unmount', '-b',
                 device.device_node], check=True)
        else:
            run(['udisksctl', 'mount', '-b',
                device.device_node], check=True)

    def get_available_mounts(self) -> List[Device]:
        """
        Returns a list of mountable partitions.

        :return: List of mountable partitions
        :rtype: List[Device]
        """
        mountable_devices = []
        for dev in self.context.list_devices(subsystem='block',
                                             DEVTYPE='partition'):
            mountable_devices.append(dev)
        return mountable_devices

    def mount_and_add_iso(self, iso_path: str):
        """
        """
        if (os.path.isfile(iso_path) and
                os.path.splitext(iso_path)[1] == '.iso'):
            result = run(['udisksctl', 'loop-setup', '-f', iso_path],
                         check=True, text=True, stdout=PIPE,
                         stderr=PIPE, universal_newlines=True).stdout.strip()
            result = str.split(result, ' ')[-1][:-1]
            device = Devices.from_device_file(self.context, result)
            self.add(device)
            self.toggle_mount(device)
            self.layoutChanged.emit()



import os
import logging
from typing import List
from tarfile import is_tarfile
from zipfile import is_zipfile
from rarfile import is_rarfile
from subprocess import CalledProcessError

from PySide6.QtWidgets import (QApplication, QFileSystemModel, QLineEdit,
                               QLabel, QMenu, QToolButton, QInputDialog,
                               QMessageBox, QMainWindow, QProgressDialog)
from PySide6.QtCore import (QFile, QDir, QFileInfo, QProcess, QStandardPaths,
                            Qt, QThreadPool)
from PySide6.QtGui import QKeySequence, QIcon

from .form import Ui_tfm

from .stack import stack
from .bookmarks_model import bookmarks_model as bm
from .mounts_model import mounts_model
from .paste_worker import paste_worker as pw
from .extract_worker import extract_worker as ew
import tfm.utility as utility

from send2trash import send2trash
from pyudev import Context, Device, Monitor, MonitorObserver


class tfm(QMainWindow, Ui_tfm):
    """
    The main window class, setting up the application and implementing the
    needed events.
    """

    def __init__(self, args: List[str]):
        """
        At the moment the very, very long initialization of the main window,
        setting up everything.

        :param default_path: Use a user defined path as entrypoint. If it's
                             empty, the home directory of the current user will
                             be used.
        :type default_path: str
        """
        super(tfm, self).__init__()
        self.setupUi(self)
        self.setWindowIcon(QIcon.fromTheme('system-file-manager'))

        self.clipboard = QApplication.clipboard()
        self.marked_to_cut = []

        self.back_stack = stack()
        self.forward_stack = stack()

        self.config_dir = os.path.join(
            QStandardPaths.writableLocation(
                QStandardPaths.ConfigLocation),
            type(self).__name__)

        print(args)
        self.current_path = utility.handle_args(args)
        self.default_path = self.current_path

        self.threadpool = QThreadPool()

        # MAIN VIEW #
        # set up QFileSystemModel
        self.filesystem = QFileSystemModel()
        self.filesystem.setRootPath(self.current_path)
        self.filesystem.setReadOnly(False)

        # connect QFileSystemModel to View
        self.table_view.setModel(self.filesystem)
        self.table_view.setRootIndex(
            self.filesystem.index(self.current_path))

        # set up header
        self.horizontal_header = self.table_view.horizontalHeader()
        self.horizontal_header.setSectionsMovable(True)
        # name
        self.horizontal_header.resizeSection(0, 200)
        # size
        self.horizontal_header.resizeSection(1, 100)
        # type
        self.horizontal_header.resizeSection(2, 100)

        # FS TREE #
        # create seperate FileSystemModel for the fs tree
        self.fs_tree_model = QFileSystemModel()
        self.fs_tree_model.setFilter(QDir.AllDirs | QDir.NoDotAndDotDot)
        self.fs_tree_model.setRootPath(QDir.rootPath())

        # connect model to view
        self.fs_tree.setModel(self.fs_tree_model)
        # hide unneeded columns
        for column in range(1, self.fs_tree.header().count()):
            self.fs_tree.hideColumn(column)
        # expand root item
        self.fs_tree.expand(self.fs_tree_model.index(0, 0))

        # BOOKMARKS #
        self.bookmarks = bm(path_to_bookmark_file=os.path.join(self.config_dir,
                                                               'bookmarks'))
        self.bookmark_view.setModel(self.bookmarks)

        # MOUNTS #
        self.udev_context = Context()
        self.mounts = mounts_model(context=self.udev_context)
        self.mounts_view.setModel(self.mounts)
        udev_monitor = Monitor.from_netlink(self.udev_context)
        udev_monitor.filter_by(subsystem='block', device_type='partition')
        udev_observer = MonitorObserver(udev_monitor,
                                        self.devices_changed)
        udev_observer.start()

        # STATUSBAR #
        # TODO: dir info
        self.item_info = QLabel()
        # self.dir_info = QLabel()
        self.part_info = QLabel()
        self.build_info = QLabel()
        self.statusbar.addPermanentWidget(self.item_info)
        # self.statusbar.addPermanentWidget(self.dir_info)
        self.statusbar.addPermanentWidget(self.part_info)
        self.statusbar.addPermanentWidget(self.build_info)
        self.build_info.setText('0.3.2 - DEV')

        self.part_info.setText(utility.part_info(self.current_path))

        # TOOLBAR #
        # initially disable back/forward navigation
        self.action_back.setEnabled(False)
        self.action_forward.setEnabled(False)

        # main menu
        self.main_menu = QMenu()
        self.main_menu.addAction(self.action_show_hidden)

        self.menu_button = QToolButton()
        self.menu_button.setMenu(self.main_menu)
        self.menu_button.setPopupMode(QToolButton.InstantPopup)
        self.menu_button.setDefaultAction(self.action_menu)

        self.toolbar.insertWidget(self.action_back, self.menu_button)

        # adress bar
        self.adressbar = QLineEdit()
        self.adressbar.setText(self.current_path)
        self.toolbar.insertWidget(self.action_go, self.adressbar)

        # menu for new file or directory
        self.new_menu = QMenu()
        self.new_menu.addAction(self.action_new_dir)
        self.new_menu.addAction(self.action_new_file)

        self.new_button = QToolButton()
        self.new_button.setMenu(self.new_menu)
        self.new_button.setPopupMode(QToolButton.MenuButtonPopup)
        self.new_button.setDefaultAction(self.action_new_dir)

        self.toolbar.insertWidget(self.action_back, self.new_button)

        self.connect_actions_to_events()
        self.set_shortcuts()
        self.set_icons()
        self.set_context_menus()

    # ---------------- setup ----------------------------------------------- #
    def set_icons(self):
        """
        Adds icons to all applicable actions.
        """
        self.action_back.setIcon(QIcon.fromTheme('go-previous'))
        self.action_forward.setIcon(QIcon.fromTheme('go-next'))
        self.action_home.setIcon(QIcon.fromTheme('go-home'))
        self.action_up.setIcon(QIcon.fromTheme('go-up'))
        self.action_go.setIcon(QIcon.fromTheme('go-jump'))
        self.action_new_dir.setIcon(QIcon.fromTheme('folder-new'))
        self.action_new_file.setIcon(QIcon.fromTheme('document-new'))
        self.action_menu.setIcon(QIcon.fromTheme('start-here'))
        self.action_copy.setIcon(QIcon.fromTheme('edit-copy'))
        self.action_cut.setIcon(QIcon.fromTheme('edit-cut'))
        self.action_paste.setIcon(QIcon.fromTheme('edit-paste'))
        self.action_delete.setIcon(QIcon.fromTheme('edit-delete'))
        self.action_add_to_bookmarks.setIcon(QIcon.fromTheme('list-add'))
        self.action_remove_bookmark.setIcon(QIcon.fromTheme('list-remove'))

    def connect_actions_to_events(self):
        """
        Connects actions to their event functions.
        """
        self.adressbar.returnPressed.connect(self.action_go_event)

        self.action_go.triggered.connect(self.action_go_event)
        self.action_home.triggered.connect(self.action_home_event)
        self.action_up.triggered.connect(self.action_up_event)
        self.action_back.triggered.connect(self.action_back_event)
        self.action_forward.triggered.connect(self.action_forward_event)
        self.action_copy.triggered.connect(self.action_copy_event)
        self.action_copy_path.triggered.connect(self.action_copy_path_event)
        self.action_paste.triggered.connect(self.action_paste_event)
        self.action_cut.triggered.connect(self.action_cut_event)
        self.action_delete.triggered.connect(self.action_delete_event)
        self.action_rename.triggered.connect(self.action_rename_event)
        self.action_show_hidden.toggled.connect(self.action_show_hidden_event)
        self.action_new_dir.triggered.connect(self.action_new_dir_event)
        self.action_new_file.triggered.connect(self.action_new_file_event)
        self.action_add_to_bookmarks.triggered.connect(
            self.action_add_to_bookmarks_event)
        self.action_remove_bookmark.triggered.connect(
            self.action_remove_bookmark_event)
        self.action_extract_here.triggered.connect(
            self.action_extract_here_event)
        self.action_mount_iso.triggered.connect(self.action_mount_iso_event)

        self.bookmark_view.pressed.connect(self.bookmark_selected_event)
        self.fs_tree.pressed.connect(self.fs_tree_event)
        self.mounts_view.pressed.connect(self.mount_selected_event)
        self.mounts_view.activated.connect(self.mount_toggle_event)
        self.table_view.pressed.connect(self.item_selected_event)
        self.table_view.activated.connect(self.item_open_event)

    def set_shortcuts(self):
        """
        Set shortcuts for actions.
        """
        self.action_copy.setShortcuts(
            QKeySequence.keyBindings(QKeySequence.Copy))
        self.action_paste.setShortcuts(
            QKeySequence.keyBindings(QKeySequence.Paste))
        self.action_cut.setShortcuts(
            QKeySequence.keyBindings(QKeySequence.Cut))
        self.action_delete.setShortcuts(
            QKeySequence.keyBindings(QKeySequence.Delete))

    def set_context_menus(self):
        """
        Setup context menus.
        """
        self.table_view.addAction(self.action_copy)
        self.table_view.addAction(self.action_copy_path)
        self.table_view.addAction(self.action_paste)
        self.table_view.addAction(self.action_cut)
        self.table_view.addAction(self.action_rename)
        self.table_view.addAction(self.action_delete)
        self.table_view.addAction(self.action_show_hidden)

        self.bookmark_view.addAction(self.action_remove_bookmark)

    # ---------------- events ---------------------------------------------- #
    def action_new_dir_event(self):
        """
        Prompts the user for a dir name and creates one with that name in the
        current directory.
        """
        dir_name, ok = QInputDialog().getText(self,
                                              'Create new directory',
                                              'Directory name:')
        if (ok):
            if (dir_name):
                if not QDir().mkpath(os.path.join(self.current_path,
                                                  dir_name)):
                    dialog = utility.message_dialog('Directory could not be'
                                                    + ' created.',
                                                    QMessageBox.Critical)
                    dialog.exec()
            else:
                dialog = utility.message_dialog('Please enter a name for the '
                                                + 'new directory.',
                                                QMessageBox.Warning)
                dialog.exec()

    def action_new_file_event(self):
        """
        Prompts the user for a file name and creates one with that name in the
        current directory.
        """
        file_name, ok = QInputDialog().getText(self,
                                               'Create new file',
                                               'File name:')
        if (ok):
            if (file_name):
                try:
                    with open(os.path.join(self.current_path, file_name), 'a'):
                        pass
                except OSError:
                    dialog = utility.message_dialog('File cannot be created.',
                                                    QMessageBox.Warning)
                    dialog.exec()
            else:
                dialog = utility.message_dialog('Please enter a name for the'
                                                + ' new file.',
                                                QMessageBox.Warning)
                dialog.exec()

    def action_go_event(self):
        """
        Changes the current dir to the one, that is given in the adress bar.
        """
        next_dir = QDir(self.adressbar.text())
        if (next_dir.isAbsolute() and next_dir.exists()):
            next_path = next_dir.path()
            self.update_current_path(next_path)
        else:
            dialog = utility.message_dialog('The path entered does not exist.',
                                            QMessageBox.Warning)
            dialog.exec()

    def action_home_event(self):
        """
        Changes the current dir to the users $HOME.
        """
        next_path = QDir().homePath()
        self.update_current_path(next_path)

    def action_up_event(self):
        """
        Changes the current dir to one dir up in the hierarchy
        """
        dir_up = QDir(self.current_path)
        if (dir_up.cdUp()):
            next_path = dir_up.path()
            self.update_current_path(next_path)
        else:
            logging.warning('There is no directory above the one you are in.'
                            + ' This message should not occur, because the'
                            + ' button triggering this should be deactivated'
                            + ' Please report this issue.')

    def action_back_event(self):
        """
        Goes back one step in the navigation history. If after this, there are
        no more steps, it disables it's button.
        """
        next_path = self.back_stack.pop()
        self.forward_stack.push(self.current_path)
        self.action_forward.setEnabled(True)
        if (self.back_stack.empty()):
            self.action_back.setEnabled(False)
        self.update_current_path(next_path,
                                 skip_stack=True,
                                 reset_forward_stack=False)

    def action_forward_event(self):
        """
        Goes forward one step in the navigation history. If after this, there
        are no more steps, it disables it's button.
        """
        if (not self.forward_stack.empty()):
            next_path = self.forward_stack.pop()
            self.update_current_path(next_path, reset_forward_stack=False)
            # disable forward action if there are no more forward actions
            if (self.forward_stack.empty()):
                self.action_forward.setEnabled(False)
        else:
            logging.warning('Forward stack unexpectedly empty. This should not'
                            + ' happen, because the button triggering this'
                            + ' message should be deactivated. Please report'
                            + ' this issue.')

    def item_open_event(self):
        """
        Opens the selected files using xdg-open, runs it, if it is executable
        or changes the current dir if it's dir.
        """
        selected_items = []
        for index in self.table_view.selectedIndexes():
            if index.column() == 0:
                selected_items.append(QFileInfo(os.path.join(self.current_path,
                                      index.siblingAtColumn(0).data())))

        # warn before accidentally open a bunch of files
        open = True
        if len(selected_items) > 3:
            dialog = utility.question_dialog('Do you really want to open '
                                             + str(len(selected_items))
                                             + ' files?')
            button = dialog.exec()
            if button == QMessageBox.Cancel:
                open = False

        if open:
            for item in selected_items:
                if (item.isDir()):
                    next_path = item.absoluteFilePath()
                    self.update_current_path(next_path)
                elif (item.isFile()):
                    if (item.isExecutable()):
                        QProcess().startDetached(item.absoluteFilePath(),
                                                 [],
                                                 self.current_path)
                    else:
                        QProcess().startDetached('xdg-open',
                                                 [item.absoluteFilePath()],
                                                 self.current_path)
                else:
                    dialog = utility.message_dialog('The type of the selected'
                                                    + ' file can not be'
                                                    + ' detected.',
                                                    QMessageBox.Warning)
                    dialog.exec()

    def item_selected_event(self):
        """
        Updates statusbar info, depending on the selected file.
        """
        selected_items = self.table_view.selectedIndexes()
        files = []
        for item in selected_items:
            files.append(os.path.join(self.current_path,
                                      item.siblingAtColumn(0).data()))
        files = list(set(files))
        if len(files) == 1:
            # update context menu according to file type
            if (os.path.isdir(files[0])):
                self.table_view.addAction(self.action_add_to_bookmarks)
                self.table_view.removeAction(self.action_extract_here)
                self.table_view.removeAction(self.action_mount_iso)
            else:
                self.table_view.removeAction(self.action_add_to_bookmarks)
                if (is_tarfile(files[0]) or is_zipfile(files[0]) or
                        is_rarfile(files[0])):
                    self.table_view.addAction(self.action_extract_here)
                    if (os.path.splitext(files[0])[1] == '.iso'):
                        self.table_view.addAction(self.action_mount_iso)
                else:
                    self.table_view.removeAction(self.action_extract_here)
                    self.table_view.removeAction(self.action_mount_iso)

        files = list(set(files))
        self.item_info.setText(utility.file_info(files))

    def fs_tree_event(self):
        """
        Changes the current dir to the dir selected in the fs_tree view.
        """
        next_path = self.fs_tree_model.filePath(self.fs_tree.currentIndex())
        self.update_current_path(next_path)

    def action_extract_here_event(self):
        """
        Extracts the given file if possible.
        """
        selected_item = os.path.join(self.current_path,
                                     self.table_view.currentIndex().
                                     siblingAtColumn(0).data())
        # setup QProgressDialog
        progress_dialog = QProgressDialog(parent=self)
        progress_dialog.reset()
        progress_dialog.setMinimumDuration(1000)
        progress_dialog.setLabelText('Extracting archive ' + selected_item
                                     + '...')
        progress_dialog.setValue(0)
        # setup extract_worker
        extract_worker = ew(archive_path=selected_item)

        # canceled
        progress_dialog.canceled.connect(extract_worker.cancel,
                                         type=Qt.DirectConnection)
        # ready
        extract_worker.signals.ready.connect(progress_dialog.setMaximum)
        # progress
        extract_worker.signals.progress.connect(progress_dialog.setValue)
        # progress message
        extract_worker.signals.progress_message.connect(
            progress_dialog.setLabelText)
        # finished
        extract_worker.signals.finished.connect(progress_dialog.reset)

        self.threadpool.start(extract_worker)

    def action_copy_event(self):
        """
        Copies the currently selected files to the clipboard.
        """
        # get current selection
        _, mime_data = utility.get_MIME(
            self.table_view.selectionModel().selectedIndexes())
        self.clipboard.setMimeData(mime_data)

    def action_paste_event(self):
        """
        Pastes the files currently in the clipboard to the current dir.
        """
        # setup QProgressDialog
        progress_dialog = QProgressDialog(parent=self)
        progress_dialog.reset()
        progress_dialog.setMinimumDuration(1000)
        progress_dialog.setLabelText('Pasting...')
        progress_dialog.setValue(0)
        # setup paste_worker
        paste_worker = pw(clipboard=self.clipboard,
                          target_path=self.current_path,
                          marked_to_cut=self.marked_to_cut)
        # canceled
        progress_dialog.canceled.connect(paste_worker.cancel,
                                         type=Qt.DirectConnection)

        # ready
        paste_worker.signals.ready.connect(progress_dialog.setMaximum)
        # progress
        paste_worker.signals.progress.connect(progress_dialog.setValue)
        # progress message
        paste_worker.signals.progress_message.connect(
            progress_dialog.setLabelText)
        # finished
        paste_worker.signals.finished.connect(progress_dialog.reset)

        self.threadpool.start(paste_worker)

    # TODO: visual feedback for cut files
    def action_cut_event(self):
        """
        Copies the current selection to the clipboard and marks them as cut.
        """
        self.marked_to_cut, mime_data = utility.get_MIME(
            self.table_view.selectionModel().selectedIndexes())
        self.clipboard.setMimeData(mime_data)

    def action_delete_event(self):
        """
        Throws the current selection in the trash after asking for
        confirmation.
        """
        path_list = utility.indexes_to_paths(
            self.table_view.selectionModel().selectedIndexes())
        if len(path_list) < 0:
            return

        if (len(path_list) == 1):
            confirmation_msg = ('Do you want to throw "'
                                + os.path.basename(path_list[0])
                                + '" in the trash?')
        else:
            confirmation_msg = ('Do you want to throw the '
                                + str(len(path_list))
                                + ' selected items and all the items contained'
                                + ' in them in the trash?')

        dialog = utility.question_dialog(confirmation_msg)
        button_clicked = dialog.exec()

        if (button_clicked == QMessageBox.Yes):
            for item in path_list:
                send2trash(item)

    def action_rename_event(self):
        """
        Prompts the user to enter a new name and changes the selected file's
        name.
        """
        file_name = self.table_view.currentIndex().data()
        new_file_name, ok = QInputDialog().getText(self,
                                                   'Rename ' + file_name,
                                                   'New name:')
        if (ok):
            if (new_file_name):
                QFile().rename(os.path.join(self.current_path, file_name),
                               os.path.join(self.current_path, new_file_name))
            else:
                dialog = utility.message_dialog('Please enter the new name '
                                                + 'for the file.',
                                                QMessageBox.Warning)
                dialog.exec()

    def action_show_hidden_event(self):
        """
        Shows hidden files and dirs in the main view.
        """
        if self.action_show_hidden.isChecked():
            self.filesystem.setFilter(QDir.AllEntries
                                      | QDir.NoDotAndDotDot
                                      | QDir.AllDirs
                                      | QDir.Hidden)
        else:
            self.filesystem.setFilter(QDir.AllEntries
                                      | QDir.NoDotAndDotDot
                                      | QDir.AllDirs)

    def action_copy_path_event(self):
        """
        Copy path of currently selected item to clipboard.
        """
        path = os.path.join(self.current_path,
                            self.table_view.currentIndex().siblingAtColumn(0)
                            .data())
        self.clipboard.setText(path)

    def action_add_to_bookmarks_event(self):
        """
        Adds the selected dir to the bookmark by the name the user gave it via
        a dialog.
        """
        path = os.path.join(self.current_path,
                            self.table_view.currentIndex().data())
        if (os.path.isdir(path)):
            name, ok = QInputDialog().getText(self,
                                              'Create new bookmark',
                                              'Bookmark name:',
                                              text=self.table_view.
                                              currentIndex().data())
            if (ok):
                if (name):
                    self.bookmarks.add_bookmark(name, path)
                    self.bookmark_view.reset()
                else:
                    dialog = utility.message_dialog('Please enter name for'
                                                    + ' the bookmark.',
                                                    QMessageBox.Warning)
                    dialog.exec()
        else:
            dialog = utility.message_dialog('Please select a directory, not'
                                            + ' a file, to create a bookmark.',
                                            QMessageBox.Warning)
            dialog.exec()

    def action_remove_bookmark_event(self):
        """
        Removes the selected bookmark after asking the user for confirmation.
        """
        name = self.bookmark_view.currentIndex().data()
        dialog = utility.question_dialog('Do you really want to delete this '
                                         + 'bookmark ('
                                         + name
                                         + ')?')
        button_clicked = dialog.exec()

        if (button_clicked == QMessageBox.Yes):
            self.bookmarks.remove_bookmark(name)
            self.bookmark_view.reset()

    def bookmark_selected_event(self):
        """
        Changes the current dir to the one associated with the selected
        bookmark.
        """
        next_path = self.bookmarks.get_path_from_name(
            self.bookmark_view.currentIndex().data())
        self.update_current_path(next_path)

    def mount_selected_event(self):
        """
        Checks the mount point of the selected drive and makes it the current
        path.
        """
        try:
            device = self.mounts.get_device_at(
                self.mounts_view.currentIndex().row())
        except IndexError:
            logging.critical('Device index out of range. This should never '
                             + 'happen, please report this error.')
            exit(1)
        mount_point = self.mounts.get_mount_point(device)
        if mount_point != '':
            self.update_current_path(mount_point)

    def action_mount_iso_event(self):
        """
        """
        iso_path = os.path.join(self.current_path,
                                self.table_view.currentIndex()
                                .siblingAtColumn(0).data())
        self.mounts.mount_and_add_iso(iso_path)

    # TODO: handle performance better
    def mount_toggle_event(self):
        """
        Tries to mount or unmount the selected drive.
        """
        try:
            device = self.mounts.get_device_at(
                self.mounts_view.currentIndex().row())
        except IndexError:
            logging.critical('Device index out of range. This should never '
                             + 'happen, please report this error.')
            exit(1)
        former_mount_point = self.mounts.get_mount_point(device)
        try:
            self.mounts.toggle_mount(device)
        except CalledProcessError:
            dialog = utility.message_dialog('udisks2 was unable to mount or'
                                            + ' unmount this devices. Please'
                                            + ' check, if it is installed'
                                            + ' correctly and report this'
                                            + ' error, if the problem'
                                            + ' presists.',
                                            QMessageBox.Critical)
            dialog.exec()
            return
        mount_point = self.mounts.get_mount_point(device)
        if mount_point != '':
            self.mount_selected_event()
        elif former_mount_point == self.current_path:
            self.update_current_path(self.default_path)

    # ---------------- functions ------------------------------------------- #
    def update_current_path(self,
                            next_path: str,
                            skip_stack=False,
                            reset_forward_stack=True):
        """
        Updates the current path to the value given in next_path. It also
        updates the UI accordingly and changes the forward an backward
        stacks, if needed.

        :param next_path: The path to the directory that should become the next
                          dir.
        :type next_path: str
        :param skip_stack: This determines wether the stack handling should be
                           skipped. Default value is False.
        :type skip_stack: bool
        :param reset_forward_stack: This determines wether the forward stack
                                    should be reset or not. Default is True
        :type reset_forward_stack: bool
        """
        self.filesystem.setRootPath(next_path)
        self.table_view.setRootIndex(
            self.filesystem.index(next_path))
        self.table_view.setCurrentIndex(
            self.filesystem.index(self.current_path))
        self.adressbar.setText(next_path)
        self.part_info.setText(utility.part_info(next_path))
        # disable up navigation if in fs root
        if (next_path == QDir().rootPath()):
            self.action_up.setEnabled(False)
        else:
            self.action_up.setEnabled(True)
        # handle back stack
        if (not skip_stack):
            if (self.back_stack.empty()
                    or self.back_stack.top() != self.current_path):
                self.back_stack.push(self.current_path)
                # reenable back navigation
                self.action_back.setEnabled(True)
        if (reset_forward_stack):
            self.forward_stack.drop()
            self.action_forward.setEnabled(False)
        self.current_path = next_path

    def devices_changed(self, action: str, device: Device):
        """
        Updates the mountable devices. Is run, when udev detects a change in
        the block device subsystem.

        :param action: Action that happened to the block device.
        :type action: str
        :param device: Device that changed it's status.
        :type device: Device
        """
        if action == 'add':
            self.mounts.add(device)
        elif action == 'remove':
            self.mounts.remove(device)
        self.mounts.layoutChanged.emit()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'form.ui'
##
## Created by: Qt User Interface Compiler version 6.2.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QHBoxLayout, QHeaderView,
    QListView, QMainWindow, QSizePolicy, QSplitter,
    QStatusBar, QTableView, QToolBar, QTreeView,
    QWidget)

class Ui_tfm(object):
    def setupUi(self, tfm):
        if not tfm.objectName():
            tfm.setObjectName(u"tfm")
        tfm.resize(800, 600)
        tfm.setContextMenuPolicy(Qt.NoContextMenu)
        tfm.setLocale(QLocale(QLocale.English, QLocale.UnitedStates))
        self.action_menu = QAction(tfm)
        self.action_menu.setObjectName(u"action_menu")
        self.action_new_dir = QAction(tfm)
        self.action_new_dir.setObjectName(u"action_new_dir")
        self.action_new_file = QAction(tfm)
        self.action_new_file.setObjectName(u"action_new_file")
        self.action_back = QAction(tfm)
        self.action_back.setObjectName(u"action_back")
        self.action_up = QAction(tfm)
        self.action_up.setObjectName(u"action_up")
        self.action_home = QAction(tfm)
        self.action_home.setObjectName(u"action_home")
        self.action_go = QAction(tfm)
        self.action_go.setObjectName(u"action_go")
        self.action_forward = QAction(tfm)
        self.action_forward.setObjectName(u"action_forward")
        self.action_copy = QAction(tfm)
        self.action_copy.setObjectName(u"action_copy")
        self.action_paste = QAction(tfm)
        self.action_paste.setObjectName(u"action_paste")
        self.action_cut = QAction(tfm)
        self.action_cut.setObjectName(u"action_cut")
        self.action_show_hidden = QAction(tfm)
        self.action_show_hidden.setObjectName(u"action_show_hidden")
        self.action_show_hidden.setCheckable(True)
        self.action_delete = QAction(tfm)
        self.action_delete.setObjectName(u"action_delete")
        self.action_rename = QAction(tfm)
        self.action_rename.setObjectName(u"action_rename")
        self.action_add_to_bookmarks = QAction(tfm)
        self.action_add_to_bookmarks.setObjectName(u"action_add_to_bookmarks")
        self.action_remove_bookmark = QAction(tfm)
        self.action_remove_bookmark.setObjectName(u"action_remove_bookmark")
        self.action_copy_path = QAction(tfm)
        self.action_copy_path.setObjectName(u"action_copy_path")
        self.action_extract_here = QAction(tfm)
        self.action_extract_here.setObjectName(u"action_extract_here")
        self.action_mount_iso = QAction(tfm)
        self.action_mount_iso.setObjectName(u"action_mount_iso")
        self.centralwidget = QWidget(tfm)
        self.centralwidget.setObjectName(u"centralwidget")
        self.horizontalLayout = QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.vsplit = QSplitter(self.centralwidget)
        self.vsplit.setObjectName(u"vsplit")
        self.vsplit.setOrientation(Qt.Horizontal)
        self.hsplit = QSplitter(self.vsplit)
        self.hsplit.setObjectName(u"hsplit")
        self.hsplit.setOrientation(Qt.Vertical)
        self.fs_tree = QTreeView(self.hsplit)
        self.fs_tree.setObjectName(u"fs_tree")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.fs_tree.sizePolicy().hasHeightForWidth())
        self.fs_tree.setSizePolicy(sizePolicy)
        self.fs_tree.setTabKeyNavigation(True)
        self.fs_tree.setProperty("showDropIndicator", True)
        self.fs_tree.setDragEnabled(False)
        self.fs_tree.setDragDropMode(QAbstractItemView.DropOnly)
        self.fs_tree.setDefaultDropAction(Qt.MoveAction)
        self.fs_tree.setRootIsDecorated(False)
        self.hsplit.addWidget(self.fs_tree)
        self.fs_tree.header().setVisible(False)
        self.bookmark_view = QListView(self.hsplit)
        self.bookmark_view.setObjectName(u"bookmark_view")
        self.bookmark_view.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.bookmark_view.setEditTriggers(QAbstractItemView.DoubleClicked|QAbstractItemView.EditKeyPressed|QAbstractItemView.SelectedClicked)
        self.bookmark_view.setTabKeyNavigation(True)
        self.bookmark_view.setDragDropOverwriteMode(True)
        self.bookmark_view.setDragDropMode(QAbstractItemView.DropOnly)
        self.bookmark_view.setDefaultDropAction(Qt.CopyAction)
        self.bookmark_view.setAlternatingRowColors(True)
        self.hsplit.addWidget(self.bookmark_view)
        self.mounts_view = QListView(self.hsplit)
        self.mounts_view.setObjectName(u"mounts_view")
        self.mounts_view.setTabKeyNavigation(True)
        self.mounts_view.setAlternatingRowColors(True)
        self.hsplit.addWidget(self.mounts_view)
        self.vsplit.addWidget(self.hsplit)
        self.table_view = QTableView(self.vsplit)
        self.table_view.setObjectName(u"table_view")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.table_view.sizePolicy().hasHeightForWidth())
        self.table_view.setSizePolicy(sizePolicy1)
        self.table_view.viewport().setProperty("cursor", QCursor(Qt.ArrowCursor))
        self.table_view.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.table_view.setAcceptDrops(True)
        self.table_view.setToolTipDuration(-3)
        self.table_view.setEditTriggers(QAbstractItemView.AnyKeyPressed|QAbstractItemView.EditKeyPressed|QAbstractItemView.SelectedClicked)
        self.table_view.setDragEnabled(True)
        self.table_view.setDragDropMode(QAbstractItemView.DragDrop)
        self.table_view.setDefaultDropAction(Qt.MoveAction)
        self.table_view.setAlternatingRowColors(True)
        self.table_view.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_view.setTextElideMode(Qt.ElideMiddle)
        self.table_view.setShowGrid(False)
        self.table_view.setSortingEnabled(True)
        self.table_view.setWordWrap(False)
        self.vsplit.addWidget(self.table_view)
        self.table_view.horizontalHeader().setDefaultSectionSize(150)
        self.table_view.horizontalHeader().setHighlightSections(False)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.verticalHeader().setVisible(False)

        self.horizontalLayout.addWidget(self.vsplit)

        tfm.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(tfm)
        self.statusbar.setObjectName(u"statusbar")
        self.statusbar.setContextMenuPolicy(Qt.NoContextMenu)
        self.statusbar.setSizeGripEnabled(False)
        tfm.setStatusBar(self.statusbar)
        self.toolbar = QToolBar(tfm)
        self.toolbar.setObjectName(u"toolbar")
        self.toolbar.setContextMenuPolicy(Qt.PreventContextMenu)
        self.toolbar.setMovable(False)
        tfm.addToolBar(Qt.TopToolBarArea, self.toolbar)

        self.toolbar.addAction(self.action_back)
        self.toolbar.addAction(self.action_forward)
        self.toolbar.addAction(self.action_up)
        self.toolbar.addAction(self.action_home)
        self.toolbar.addAction(self.action_go)

        self.retranslateUi(tfm)

        QMetaObject.connectSlotsByName(tfm)
    # setupUi

    def retranslateUi(self, tfm):
        tfm.setWindowTitle(QCoreApplication.translate("tfm", u"tfm", None))
        self.action_menu.setText(QCoreApplication.translate("tfm", u"Menu", None))
        self.action_new_dir.setText(QCoreApplication.translate("tfm", u"New Directory", None))
        self.action_new_file.setText(QCoreApplication.translate("tfm", u"New File", None))
        self.action_back.setText(QCoreApplication.translate("tfm", u"Back", None))
        self.action_up.setText(QCoreApplication.translate("tfm", u"Up", None))
        self.action_home.setText(QCoreApplication.translate("tfm", u"Home", None))
        self.action_go.setText(QCoreApplication.translate("tfm", u"Go", None))
#if QT_CONFIG(tooltip)
        self.action_go.setToolTip(QCoreApplication.translate("tfm", u"Go or Reload", None))
#endif // QT_CONFIG(tooltip)
        self.action_forward.setText(QCoreApplication.translate("tfm", u"Forward", None))
        self.action_copy.setText(QCoreApplication.translate("tfm", u"Copy", None))
        self.action_paste.setText(QCoreApplication.translate("tfm", u"Paste", None))
        self.action_cut.setText(QCoreApplication.translate("tfm", u"Cut", None))
        self.action_show_hidden.setText(QCoreApplication.translate("tfm", u"Show hidden files", None))
#if QT_CONFIG(shortcut)
        self.action_show_hidden.setShortcut(QCoreApplication.translate("tfm", u"Ctrl+H", None))
#endif // QT_CONFIG(shortcut)
        self.action_delete.setText(QCoreApplication.translate("tfm", u"Delete", None))
        self.action_rename.setText(QCoreApplication.translate("tfm", u"Rename", None))
        self.action_add_to_bookmarks.setText(QCoreApplication.translate("tfm", u"Add to bookmarks", None))
        self.action_remove_bookmark.setText(QCoreApplication.translate("tfm", u"Remove bookmark", None))
#if QT_CONFIG(tooltip)
        self.action_remove_bookmark.setToolTip(QCoreApplication.translate("tfm", u"Remove bookmark", None))
#endif // QT_CONFIG(tooltip)
        self.action_copy_path.setText(QCoreApplication.translate("tfm", u"Copy path", None))
        self.action_extract_here.setText(QCoreApplication.translate("tfm", u"Extract here", None))
#if QT_CONFIG(tooltip)
        self.action_extract_here.setToolTip(QCoreApplication.translate("tfm", u"Extract here", None))
#endif // QT_CONFIG(tooltip)
        self.action_mount_iso.setText(QCoreApplication.translate("tfm", u"Mount ISO", None))
#if QT_CONFIG(tooltip)
        self.action_mount_iso.setToolTip(QCoreApplication.translate("tfm", u"Mount ISO", None))
#endif // QT_CONFIG(tooltip)
    # retranslateUi




import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='tfm',
    version='0.3.3-ALPHA',
    packages=['tfm', ],
    license='GPLv3',
    author='Till Mahlburg',
    # author_email="author@example.com",
    description='A simple and Qt file manager for Linux',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tmahlburg/tfm',
    install_requires=[
        'pyside6',
        'prefixed',
        'send2trash',
        'pyudev',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: X11 Applications :: Qt',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3'
        + ' (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Desktop Environment :: File Managers',
    ],
    python_requires='>=3.7',
)



import os
import tarfile
import zipfile
import rarfile

from PySide6.QtCore import QRunnable, QObject, QFileInfo, Signal, Slot


class extract_signals(QObject):
    """
    QObject class to hold the signals for the extract worker, which can't do
    that, because it a subclass of QRunnable.

    Signals:
    **finished**: is emitted, when the worker is done.
    **started**: is emitted, when the worker starts.
    **ready**: is emitted, when the archive is opened and contains the number
               of files in it are known.
    **progress**: is emitted, whenever a file is done being extracted from the
                  archive, contains the number of files extracted so far.
    **progress_message**: is emitted, whenever **progress** is emitted, but is
                          a formatted string containing the information about
                          what is happening.
    """

    finished = Signal()
    started = Signal()
    ready = Signal(int)
    progress_message = Signal(str)
    progress = Signal(int)


class extract_worker(QRunnable):
    """
    Worker class, that is used to extract archives in a different thread.
    """

    def __init__(self,
                 *args,
                 archive_path: str,
                 **kwargs):
        """
        Calls QObject constructor and takes the path of the archive.

        :param archive_path: Path, where the archive is located at.
        :type archive_path: str
        """
        super(extract_worker, self).__init__(*args, **kwargs)
        self.archive_path = archive_path
        self.signals = extract_signals()

        self.is_canceled = False
        self.maximum = 0

    def progress_str(self, file_count: int) -> str:
        return ('Extracting file ' + str(file_count) + ' of ' +
                str(self.maximum) + '...')

    @Slot()
    def cancel(self):
        self.is_canceled = True

    @Slot()
    def run(self):
        """
        Main logic. Extracts the archive to a directory with the same name.
        """
        self.signals.started.emit()

        # define target directory
        target_dir = QFileInfo(self.archive_path)
        target_dir = os.path.join(target_dir.path(),
                                  target_dir.completeBaseName())

        while (os.path.isdir(target_dir)):
            target_dir += '_'

        if tarfile.is_tarfile(self.archive_path):
            with tarfile.TarFile(self.archive_path, 'r') as archive:
                self.maximum = len(archive.getnames())
                self.signals.ready.emit(self.maximum)
                file_count = 0
                self.signals.progress.emit(file_count)
                self.signals.progress_message.emit(self.progress_str(
                                                   file_count))
                for file in archive.get_members():
                    if self.is_canceled:
                        break
                    archive.extract(file, target_dir)
                    file_count += 1
                    self.signals.progress.emit(file_count)
                    self.signals.progress_message.emit(self.progress_str(
                                                       file_count))

        elif zipfile.is_zipfile(self.archive_path):
            with zipfile.ZipFile(self.archive_path, 'r') as archive:
                self.maximum = len(archive.namelist())
                self.signals.ready.emit(self.maximum)
                file_count = 0
                self.signals.progress.emit(file_count)
                self.signals.progress_message.emit(self.progress_str(
                                                   file_count))
                for file in archive.infolist():
                    if self.is_canceled:
                        break
                    archive.extract(file, target_dir)
                    file_count += 1
                    self.signals.progress.emit(file_count)
                    self.signals.progress_message.emit(self.progress_str(
                                                       file_count))

        elif rarfile.is_rarfile(self.archive_path):
            with rarfile.RarFile(self.archive_path, 'r') as archive:
                self.maximum = len(archive.namelist())
                self.signals.ready.emit(self.maximum)
                file_count = 0
                self.signals.progress.emit(file_count)
                self.signals.progress_message.emit(self.progress_str(
                                                   file_count))
                for file in archive.infolist():
                    if self.is_canceled:
                        break
                    archive.extract(file, target_dir)
                    file_count += 1
                    self.signals.progress.emit(file_count)
                    self.signals.progress_message.emit(self.progress_str(
                                                       file_count))

        self.signals.finished.emit()
